# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

import itertools

from ripe.atlas.cousteau import MeasurementRequest

from .base import Command as BaseCommand, TabularFieldsMixin
from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from ..helpers.validators import ArgumentType


class Command(TabularFieldsMixin, BaseCommand):

    NAME = "measurement-search"
    LIMITS = (1, 1000)

    STATUS_SPECIFIED = 0
    STATUS_SCHEDULED = 1
    STATUS_ONGOING = 2
    STATUS_STOPPED = 4
    STATUS_FORCED_STOP = 5
    STATUS_NO_SUITABLE_PROBES = 6
    STATUS_FAILED = 7
    STATUSES = {
        "scheduled": (STATUS_SPECIFIED, STATUS_SCHEDULED,),
        "ongoing": (STATUS_ONGOING,),
        "stopped": (
            STATUS_STOPPED,
            STATUS_FORCED_STOP,
            STATUS_NO_SUITABLE_PROBES,
            STATUS_FAILED,
        )
    }

    # Column name: (alignment, width)
    COLUMNS = {
        "id": ("<", 7),
        "type": ("<", 10),
        "description": ("<", 42),
        "status": (">", 18),
        "target": ("<", 45),
        "url": ("<", 45),
    }

    DESCRIPTION = (
        "Fetch and print measurements fulfilling specified criteria based "
        "on given filters"
    )

    def add_arguments(self):

        self.parser.add_argument(
            "--search",
            type=str,
            help="A search string.  This could match the target or description."
        )
        self.parser.add_argument(
            "--status",
            type=str,
            choices=self.STATUSES.keys(),
            help="The measurement status."
        )
        self.parser.add_argument(
            "--af",
            type=int,
            choices=(4, 6),
            help="The address family."
        )
        self.parser.add_argument(
            "--type",
            type=str,
            choices=("ping", "traceroute", "dns", "sslcert", "ntp", "http"),
            help="The measurement type."
        )
        self.parser.add_argument(
            "--field",
            type=str,
            action="append",
            choices=self.COLUMNS.keys(),
            default=[],
            help="The field(s) to display. Invoke multiple times for multiple "
                 "fields. The default is id, type, description, and status."
        )
        self.parser.add_argument(
            "--ids-only",
            action="store_true",
            default=False,
            help="Display a list of measurement ids matching your filter "
                 "criteria."
        )

        timing = self.parser.add_argument_group("Timing")
        for position in ("started", "stopped"):
            for chrono in ("before", "after"):
                timing.add_argument(
                    "--{}-{}".format(position, chrono),
                    type=ArgumentType.datetime,
                    help="Filter for measurements that {} {} a specific date. "
                         "The format required is YYYY-MM-DDTHH:MM:SS".format(
                             position, chrono)
                )

        self.parser.add_argument(
            "--limit",
            type=ArgumentType.integer_range(self.LIMITS[0], self.LIMITS[1]),
            default=50,
            help="The number of measurements to return.  The number must be "
                 "between {} and {}".format(self.LIMITS[0], self.LIMITS[1])
        )

    def run(self):

        if not self.arguments.field:
            self.arguments.field = ("id", "type", "description", "status")

        filters = self._get_filters()
        measurements = MeasurementRequest(
            return_objects=True, user_agent=self.user_agent, **filters)
        truncated_measurements = itertools.islice(
            measurements, self.arguments.limit)

        if self.arguments.ids_only:
            for measurement in truncated_measurements:
                print(measurement.id)
            return

        hr = self._get_horizontal_rule()

        print(self._get_filter_display(filters))
        print(self._get_header())
        print(colourise(hr, "bold"))

        for measurement in truncated_measurements:
            lformat = self._get_line_format().format(*self._get_line_items(measurement))
            lcolour = self._get_colour_from_status(measurement.status_id)
            log = colourise(lformat, lcolour).encode("utf8")
            print(log)

        print(colourise(hr, "bold"))

        # Print total count of found measurements
        print(("{:>" + str(len(hr)) + "}\n").format(
            "Showing {} of {} total measurements".format(
                min(self.arguments.limit, measurements.total_count),
                measurements.total_count
            )
        ))

    def _get_line_items(self, measurement):

        r = []

        for field in self.arguments.field:
            if field == "url":
                r.append("https://atlas.ripe.net/measurements/{}/".format(
                    measurement.id
                ))
            elif field == "type":
                r.append(measurement.type.lower())
                continue
            elif field == "target":
                r.append(sanitise(
                    measurement.target or
                    measurement.target_ip or
                    "-"
                )[:self.COLUMNS["target"][1]])
            elif field == "description":
                description = sanitise(measurement.description) or ""
                r.append(description[:self.COLUMNS["description"][1]])
            else:
                r.append(sanitise(getattr(measurement, field)))

        return r

    def _get_filters(self):

        r = {}

        if self.arguments.search:
            r["search"] = self.arguments.search
        if self.arguments.status:
            r["status__in"] = self.STATUSES[self.arguments.status]
        if self.arguments.af:
            r["af"] = self.arguments.af
        if self.arguments.type:
            r["type"] = self.arguments.type
        if self.arguments.started_before:
            r["start_time__lt"] = self.arguments.started_before
        if self.arguments.started_after:
            r["start_time__gt"] = self.arguments.started_after
        if self.arguments.stopped_before:
            r["stop_time__lt"] = self.arguments.stopped_before
        if self.arguments.stopped_after:
            r["stop_time__gt"] = self.arguments.stopped_after

        return r

    def _get_colour_from_status(self, status):
        if status in self.STATUSES["ongoing"]:
            return "green"
        if status == self.STATUS_STOPPED:
            return "yellow"
        if status in self.STATUSES["stopped"]:
            return "red"
        if status in self.STATUSES["scheduled"]:
            return "blue"
        return "white"



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..ipdetails import IP
from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_TRACEROUTE]

    DEFAULT_RADIUS = 2

    @staticmethod
    def add_arguments(parser):
        group = parser.add_argument_group(
            title="Optional arguments for traceroute_aspath renderer"
        )
        group.add_argument(
            "--traceroute-aspath-radius",
            type=int,
            help="Number of different ASs starting from the end of the "
                 "traceroute path. "
                 "Default: {}.".format(Renderer.DEFAULT_RADIUS),
            metavar="RADIUS",
            default=Renderer.DEFAULT_RADIUS
        )

    def __init__(self, *args, **kwargs):
        BaseRenderer.__init__(self, *args, **kwargs)
        self.paths = {}

        # Number of different ASs starting from the end of the traceroute path.
        if "arguments" in kwargs:
            self.RADIUS = kwargs["arguments"].traceroute_aspath_radius
        else:
            self.RADIUS = Renderer.DEFAULT_RADIUS

    @staticmethod
    def _get_asns_for_output(asns, radius):
        asns_with_padding = [""] * radius + asns
        asns_with_padding = asns_with_padding[-radius:]
        return " ".join(
            ["{:>8}".format("AS{}".format(asn) if asn else "") for asn in asns_with_padding]
        )

    def on_start(self):
        return "For each traceroute path toward the target, the " \
               "last {} ASNs will be shown\n\n".format(self.RADIUS)

    def on_result(self, result):

        ip_hops = []

        for hop in result.hops:
            for packet in hop.packets:
                if packet.origin:
                    ip_hops.append(packet.origin)
                    break

        asns = []

        # starting from the last hop's IP, get up to <RADIUS> ASNs
        for address in reversed(ip_hops):
            ip = IP(address)
            if ip.asn and ip.asn not in asns:
                asns.append(ip.asn)
            if len(asns) == self.RADIUS:
                break

        as_path = self._get_asns_for_output(list(reversed(asns)), self.RADIUS)

        if as_path not in self.paths:
            self.paths[as_path] = {}
            self.paths[as_path]['cnt'] = 0
            self.paths[as_path]['responded'] = 0
        self.paths[as_path]['cnt'] += 1
        if result.destination_ip_responded:
            self.paths[as_path]['responded'] += 1

        return "Probe #{:<5}: {}, {}completed\n".format(
            result.probe_id, as_path,
            "NOT " if not result.destination_ip_responded else ""
        )

    def additional(self, results):
        s = "\nNumber of probes for each AS path:\n\n"

        for as_path in self.paths:
            s += "  {}: {} probe{}, {} completed\n".format(
                as_path,
                self.paths[as_path]['cnt'],
                "s" if self.paths[as_path]['cnt'] > 1 else "",
                self.paths[as_path]['responded']
            )

        return s



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import six
COLOURS_AVAILABLE = False
try:
    # We use curses to detect ANSI colour support
    import curses
except ImportError:
    # Curses isn't available on all platforms
    try:
        import colorama  # Colorama wraps stdout/stderr on Windows
    except ImportError:
        pass
    else:
        colorama.init()
        COLOURS_AVAILABLE = True
else:
    if sys.stdout.isatty():
        curses.setupterm()
        COLOURS_AVAILABLE = curses.tigetnum("colors") >= 8


class Colour(object):

    @classmethod
    def _colourise(cls, text, colour):
        # Make sure we don't mix unicode and strings
        if six.PY2 and isinstance(text, six.string_types):
            text = unicode(text)

        return u"{}[{}m{}{}[0m".format(chr(0x1b), colour, text, chr(0x1b))

    @classmethod
    def black(cls, text):
        return cls._colourise(text, 30)

    @classmethod
    def red(cls, text):
        return cls._colourise(text, 31)

    @classmethod
    def green(cls, text):
        return cls._colourise(text, 32)

    @classmethod
    def yellow(cls, text):
        return cls._colourise(text, 33)

    @classmethod
    def blue(cls, text):
        return cls._colourise(text, 34)

    @classmethod
    def mangenta(cls, text):
        return cls._colourise(text, 35)

    @classmethod
    def cyan(cls, text):
        return cls._colourise(text, 36)

    @classmethod
    def white(cls, text):
        return cls._colourise(text, 37)

    @classmethod
    def bold(cls, text):
        return cls._colourise(text, 1)


def colourise(text, colour, fileobj=sys.stdout):
    """
    Return an ANSI escaped string of the specified content and colour, or
    the input text if colour support is not available or not appropriate.

    `fileobj` is used to determine whether the output is a terminal.
    """
    if COLOURS_AVAILABLE and fileobj.isatty():
        return getattr(Colour, colour)(text)
    else:
        return text



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ripe.atlas.cousteau import Measurement
from ripe.atlas.cousteau.exceptions import APIResponseError

from ..exceptions import RipeAtlasToolsException
from ..renderers import Renderer
from ..streaming import Stream, CaptureLimitExceeded
from .base import Command as BaseCommand
from ..helpers.validators import ArgumentType


class Command(BaseCommand):

    NAME = "stream"

    DESCRIPTION = (
        "Output the results of a public measurement as they become available"
    )
    EXTRA_DESCRIPTION = "Streaming of non-public measurements is not supported."
    URLS = {
        "detail": "/api/v2/measurements/{0}.json",
    }

    def add_arguments(self):
        self.parser.add_argument(
            "measurement_id",
            type=ArgumentType.msm_id_or_name(),
            help="The measurement id or alias you want streamed"
        )
        self.parser.add_argument(
            "--limit",
            type=int,
            help="The maximum number of results you want to stream"
        )
        self.parser.add_argument(
            "--renderer",
            choices=Renderer.get_available(),
            help="The renderer you want to use. If this isn't defined, an "
                 "appropriate renderer will be selected."
        )

        Renderer.add_arguments_for_available_renderers(self.parser)

    def run(self):

        try:
            measurement = Measurement(
                id=self.arguments.measurement_id, user_agent=self.user_agent,
            )
        except APIResponseError as e:
            raise RipeAtlasToolsException(e.args[0])

        try:
            Stream(capture_limit=self.arguments.limit).stream(
                self.arguments.renderer,
                self.arguments,
                measurement.type.lower(),
                self.arguments.measurement_id
            )
        except (KeyboardInterrupt, CaptureLimitExceeded):
            self.ok("Disconnecting from the stream")



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..helpers.sanitisers import sanitise
from .base import Renderer as BaseRenderer

THRESHOLD = 80  # %


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_SSLCERT]

    def __init__(self, *args, **kwargs):
        BaseRenderer.__init__(self, *args, **kwargs)
        self.uniqcerts = {}
        self.blob_list = []

    def additional(self, results):
        self.gather_unique_certs(results)
        most_seen_cert = self.get_nprobes_ofpopular_cert()
        for cert_id in sorted(
            self.uniqcerts,
            key=lambda pk: self.uniqcerts[pk]["cnt"],
            reverse=True
        ):
            self.blob_list.append(self.render_certificate(cert_id))
            if self.uniqcerts[cert_id]["cnt"] < most_seen_cert * THRESHOLD / 100:
                self.blob_list.extend(self.render_below_threshold(cert_id))

        return "\n".join(self.blob_list)

    def gather_unique_certs(self, results):
        for result in results:
            self.bucketize_result_cert(result)

    def bucketize_result_cert(self, result):
        for certificate in result.certificates:
            cert_id = certificate.checksum_sha256
            if cert_id not in self.uniqcerts:
                self.uniqcerts[cert_id] = {
                    "cert": None,
                    "cnt": 0,
                    "probes": []
                }
            self.uniqcerts[cert_id]["cert"] = certificate
            self.uniqcerts[cert_id]["cnt"] += 1
            self.uniqcerts[cert_id]["probes"].append(result.probe)

    def get_nprobes_ofpopular_cert(self):
        """
        Gets the number of probes that have seen the most popular
        (in terms of probes) cert.
        """
        return max(
            [self.uniqcerts[cert_id]["cnt"] for cert_id in self.uniqcerts]
        )

    def render_certificate(self, cert_id):
        """Renders the specific certificate"""
        certificate = self.uniqcerts[cert_id]["cert"]

        return self.render(
            "reports/ssl_consistency.txt",
            issuer_c=sanitise(certificate.issuer_c),
            issuer_o=sanitise(certificate.issuer_o),
            issuer_cn=sanitise(certificate.issuer_cn),
            subject_c=sanitise(certificate.subject_c),
            subject_o=sanitise(certificate.subject_o),
            subject_cn=sanitise(certificate.subject_cn),
            sha256fp=certificate.checksum_sha256,
            seenby=self.uniqcerts[cert_id]["cnt"],
            s="s" if self.uniqcerts[cert_id]["cnt"] > 1 else ""
        )

    def render_below_threshold(self, cert_id):
        """
        Print information about the given cert that is below our threshold
        of visibility.
        """
        blob_list = [
            "  Below the threshold ({0}%)".format(THRESHOLD),
            "  Probes that saw it: ",
        ]

        for probe in self.uniqcerts[cert_id]["probes"]:
            log = (
                "    ID: {id}, country code: {cc}, ASN (v4/v6): {asn4}/{asn6}"
            ).format(
                id=probe.id, cc=probe.country_code,
                asn4=probe.asn_v4, asn6=probe.asn_v6
            )
            blob_list.append(log)

        return blob_list

    def on_result(self, result):
        return ""



# Copyright (c) 2017 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

from pytz import timezone
from ripe.atlas.sagan import Result
from ripe.atlas.tools.renderers.dns_compact import Renderer

def get_fake_localzone():
    return timezone('UTC')

@mock.patch('ripe.atlas.tools.renderers.dns_compact.get_localzone', get_fake_localzone)
class TestDnsCompact(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.basic = Result.get('{"lts":92,"from":"195.113.83.16","msm_id":9211416,"fw":4790,"timestamp":1503927916,"resultset":[{"lts":92,"src_addr":"195.113.83.16","af":4,"submax":3,"proto":"UDP","subid":1,"result":{"abuf":"fViBgAABAAIAAgAEBTF4YmV0A2NvbQAAAQABwAwAAQABAAAAGQAEvmnCOsAMAAEAAQAAABkABL550oXADAACAAEAAVUJABUEZGFuYQJucwpjbG91ZGZsYXJlwBLADAACAAEAAVUJAAcEcGV0ZcBMwEcAAQABAAFVCQAErfU6acBHABwAAQABVQkAECQAywAgSQABAAAAAK31OmnAaAABAAEAAVUJAASt9TuIwGgAHAABAAFVCQAQJADLACBJAAEAAAAArfU7iA==","rt":6.45,"NSCOUNT":2,"QDCOUNT":1,"ANCOUNT":2,"ARCOUNT":4,"ID":32088,"size":199},"time":1503927916,"dst_addr":"195.113.83.55"},{"lts":93,"src_addr":"195.113.83.16","af":4,"submax":3,"proto":"UDP","subid":2,"result":{"abuf":"D/uBBQABAAAAAAAABTF4YmV0A2NvbQAAAQAB","rt":5.798,"NSCOUNT":0,"QDCOUNT":1,"ANCOUNT":0,"ARCOUNT":0,"ID":4091,"size":27},"time":1503927917,"dst_addr":"147.231.12.1"},{"lts":94,"src_addr":"2001:718:1e06::16","af":6,"submax":3,"proto":"UDP","subid":3,"result":{"abuf":"pqqBgAABAAIAAgAEBTF4YmV0A2NvbQAAAQABwAwAAQABAAAAFwAEvnnShcAMAAEAAQAAABcABL5pwjrADAACAAEAAVUHABUEcGV0ZQJucwpjbG91ZGZsYXJlwBLADAACAAEAAVUHAAcEZGFuYcBMwGgAAQABAAFVBwAErfU6acBoABwAAQABVQcAECQAywAgSQABAAAAAK31OmnARwABAAEAAVUHAASt9TuIwEcAHAABAAFVBwAQJADLACBJAAEAAAAArfU7iA==","rt":5.684,"NSCOUNT":2,"QDCOUNT":1,"ANCOUNT":2,"ARCOUNT":4,"ID":42666,"size":199},"time":1503927918,"dst_addr":"2001:718:1e06::55"}],"prb_id":4062,"group_id":9211416,"type":"dns","msm_name":"Tdig"}')
        self.noerrornodata = Result.get('{"lts":42,"from":"2001:718:1:a100::161:50","msm_id":9386425,"fw":4780,"proto":"UDP","af":6,"msm_name":"Tdig","prb_id":6068,"result":{"abuf":"8RCAgAABAAAAAQAACGlwdjRvbmx5BGFycGEAABwAAcAMAAYAAQAABikALQNzbnMDZG5zBWljYW5uA29yZwADbm9jwC94Ob7xAAAcIAAADhAACTqAAAAOEA==","rt":153.425,"NSCOUNT":1,"QDCOUNT":1,"ID":61712,"ARCOUNT":0,"ANCOUNT":0,"size":88},"timestamp":1506557387,"src_addr":"2001:718:1:a100::161:50","group_id":9386425,"type":"dns","dst_addr":"2001:4860:4860::6464"}')
        self.noresponse = Result.get('{"lts":11,"from":"2a01:538:1:f000:fa1a:67ff:fe4d:7f1d","msm_id":9386425,"fw":4780,"timestamp":1506681497,"proto":"UDP","msm_name":"Tdig","prb_id":11879,"af":6,"error":{"timeout":5000},"src_addr":"2a01:538:1:f000:fa1a:67ff:fe4d:7f1d","group_id":9386425,"type":"dns","dst_addr":"2001:4860:4860::6464"}')
        self.noabuf = Result.get('{"lts":27,"from":"80.92.240.37","msm_id":9211416,"fw":4780,"timestamp":1503927938,"resultset":[{"lts":27,"src_addr":"192.168.254.254","af":4,"submax":2,"proto":"UDP","subid":1,"time":1503927938,"error":{"timeout":5000},"dst_addr":"80.92.240.6"}],"prb_id":30410,"group_id":9211416,"type":"dns","msm_name":"Tdig"}')

    def test_basic(self):
        self.assertEqual(
            Renderer().on_result(self.basic),
            "Probe  #4062: 2017-08-28 13:45:16 NOERROR qr ra rd 1xbet.com. 25 A 190.105.194.58; 1xbet.com. 25 A 190.121.210.133\n"
            "Probe  #4062: 2017-08-28 13:45:16 REFUSED qr rd\n"
            "Probe  #4062: 2017-08-28 13:45:16 NOERROR qr ra rd 1xbet.com. 23 A 190.121.210.133; 1xbet.com. 23 A 190.105.194.58\n"

        )

    def test_noerrornodata(self):
        self.assertEqual(
            Renderer().on_result(self.noerrornodata),
            "Probe  #6068: 2017-09-28 00:09:47 NOERROR qr ra\n"
        )

    def test_noresponse(self):
        self.assertEqual(
            Renderer().on_result(self.noresponse),
            "Probe #11879: 2017-09-29 10:38:17 No response found\n"
        )

    def test_noabuf(self):
        self.assertEqual(
            Renderer().on_result(self.noabuf),
            "Probe #30410: 2017-08-28 13:45:38 No abuf found\n"
        )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..cache import cache

from ripe.atlas.cousteau import ProbeRequest
from ripe.atlas.cousteau import Probe as CProbe


class Probe(object):
    """
    A crude representation of the data we get from the API via Cousteau
    """

    EXPIRE_TIME = 60 * 60 * 24 * 30

    @classmethod
    def get(cls, pk):
        """
        Given a single id, attempt to fetch a probe object from the cache.  If
        that fails, do an API call to get it.  Don't use this for multiple
        probes unless you know they're all in the cache, or you'll be in for a
        long wait.
        """
        r = cache.get("probe:{}".format(pk))
        if not r:
            probe = CProbe(id=pk)
            cache.set("probe:{}".format(probe.id), probe, cls.EXPIRE_TIME)
            return probe

    @classmethod
    def get_many(cls, ids):
        """
        Given a list of ids, attempt to get probe objects out of the local
        cache.  Probes that cannot be found will be fetched from the API and
        cached for future use.
        """

        r = []

        fetch_ids = []
        for pk in ids:
            probe = cache.get("probe:{}".format(pk))
            if probe:
                r.append(probe)
            else:
                fetch_ids.append(str(pk))

        if fetch_ids:
            kwargs = {"id__in": fetch_ids}
            for probe in [p for p in ProbeRequest(return_objects=True, **kwargs)]:
                cache.set("probe:{}".format(probe.id), probe, cls.EXPIRE_TIME)
                r.append(probe)

        return r



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from contextlib import contextmanager
try:
    from cStringIO import StringIO
except ImportError:  # Python 3
    from io import StringIO


class FakeTTY(object):
    """
    Basic simulation of a user terminal.
    """
    def __init__(self, file_obj):
        self.file_obj = file_obj

    def __getattr__(self, name):
        return getattr(self.file_obj, name)

    def isatty(self):
        return True


@contextmanager
def capture_sys_output(use_fake_tty=False):
    """
    Wrap a block with this, and it'll capture standard out and standard error
    into handy variables:

      with capture_sys_output() as (stdout, stderr):
          self.cmd.run()

    More info: https://stackoverflow.com/questions/18651705/
    """

    capture_out, capture_err = StringIO(), StringIO()
    current_out, current_err = sys.stdout, sys.stderr
    current_in = sys.stdin
    try:
        if use_fake_tty:
            sys.stdin = FakeTTY(current_in)
        sys.stdout, sys.stderr = capture_out, capture_err
        yield capture_out, capture_err
    finally:
        sys.stdout, sys.stderr = current_out, current_err
        sys.stdin = current_in



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from collections import namedtuple

from ripe.atlas.tools.aggregators.base import (
    aggregate, ValueKeyAggregator, RangeKeyAggregator
)


class TestAggregators(unittest.TestCase):
    def setUp(self):
        self.results = []
        self.probes = []
        self.Result = namedtuple('Result', 'id probe rtt source prefix')
        self.Probe = namedtuple('Probe', 'id country asn status')
        probes = [
            (1, "GR", 333, "Connected"),
            (2, "NL", 334, "Connected"),
            (3, "SE", 335, "Connected"),
            (4, "SE", 336, "DisConnected"),
            (5, "SE", 337, "DisConnected"),
            (6, "SE", 335, "DisConnected"),
            (7, "IN", 335, "Connected"),
            (8, "DE", 338, "Connected"),
            (9, "DK", 348, "Connected"),
            (10, "DE", 338, "NeverConnected"),
            (11, "DE", 340, "DisConnected"),
        ]
        results = [
            (1, 3, "127.0.0.1", "192/8"),
            (2, 34.0, "127.0.0.1", "192/8"),
            (3, 35.0, "127.0.0.1", "193/8"),
            (4, 6, "127.0.0.1", "193/8"),
            (5, 17, "127.0.0.1", "192/8"),
            (6, 15, "127.0.0.1", "195/8"),
            (7, 35, "127.0.0.1", "192/8"),
            (8, 28, "127.0.0.1", "192/8"),
            (9, 48, "127.0.0.1", "195/8"),
            (10, 28, "127.0.0.1", "194/8"),
            (11, 40, "127.0.0.1", "193/8"),
        ]
        for index, result in enumerate(results):
            prb = self.Probe(
                id=probes[index][0],
                country=probes[index][1],
                asn=probes[index][2],
                status=probes[index][3]
            )
            r = self.Result(
                id=result[0],
                probe=prb,
                rtt=result[1],
                source=result[2],
                prefix=result[3]
            )
            self.results.append(r)
            self.probes.append(prb)

    def test_value_aggregation(self):
        """Test 2 tier aggregation with object attribute as key."""
        keys = [ValueKeyAggregator(key='probe.country'), ValueKeyAggregator(key='rtt')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'COUNTRY: DE': {
                'RTT: 28': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8'), self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')],
                'RTT: 40': [self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')]
            },
            'COUNTRY: DK': {'RTT: 48': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')]},
            'COUNTRY: GR': {'RTT: 3': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')]},
            'COUNTRY: IN': {'RTT: 35': [self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')]},
            'COUNTRY: NL': {'RTT: 34.0': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8')]},
            'COUNTRY: SE': {
                'RTT: 6': [self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')],
                'RTT: 15': [self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')],
                'RTT: 17': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8')],
                'RTT: 35.0': [self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8')]
            }
        }
        self.assertEquals(buckets, expected_output)

    def test_value_aggregation1(self):
        """Test 1 tier aggregation"""
        keys = [ValueKeyAggregator(key='probe.country')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'COUNTRY: DE': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8'), self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8'), self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')],
            'COUNTRY: DK': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')],
            'COUNTRY: GR': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')],
            'COUNTRY: IN': [self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')],
            'COUNTRY: NL': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8')],
            'COUNTRY: SE': [self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8'), self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8'), self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8'), self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')]
        }
        self.assertEquals(buckets, expected_output)

    def test_value_aggregation2(self):
        """Test 2 tier aggregation."""
        keys = [ValueKeyAggregator(key='prefix'), ValueKeyAggregator(key='rtt')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'PREFIX: 192/8': {
                'RTT: 3': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')],
                'RTT: 17': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8')],
                'RTT: 28': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8')],
                'RTT: 34.0': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8')],
                'RTT: 35': [self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')]
            },
            'PREFIX: 193/8': {
                'RTT: 6': [self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')],
                'RTT: 35.0': [self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8')],
                'RTT: 40': [self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')]
            },
            'PREFIX: 194/8': {
                'RTT: 28': [self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')]
            },
            'PREFIX: 195/8': {
                'RTT: 15': [self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')],
                'RTT: 48': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')]
            }
        }
        self.assertEquals(buckets, expected_output)

    def test_value_aggregation3(self):
        """Test 3 tier aggregation."""
        keys = [ValueKeyAggregator(key='prefix'), ValueKeyAggregator(key='rtt'), ValueKeyAggregator(key='source')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'PREFIX: 192/8': {
                'RTT: 3': {
                    'SOURCE: 127.0.0.1': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')]
                },
                'RTT: 17': {
                    'SOURCE: 127.0.0.1': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8')]
                },
                'RTT: 28': {
                    'SOURCE: 127.0.0.1': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8')]
                },
                'RTT: 34.0': {
                    'SOURCE: 127.0.0.1': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8')]
                },
                'RTT: 35': {
                    'SOURCE: 127.0.0.1': [self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')]
                }
            },
            'PREFIX: 193/8': {
                'RTT: 6': {
                    'SOURCE: 127.0.0.1': [self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')]
                },
                'RTT: 35.0': {
                    'SOURCE: 127.0.0.1': [self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8')]
                },
                'RTT: 40': {
                    'SOURCE: 127.0.0.1': [self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')]
                }
            },
            'PREFIX: 194/8': {
                'RTT: 28': {
                    'SOURCE: 127.0.0.1': [self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')]
                }
            },
            'PREFIX: 195/8': {
                'RTT: 15': {
                    'SOURCE: 127.0.0.1': [self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')]
                },
                'RTT: 48': {
                    'SOURCE: 127.0.0.1': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')]
                }
            }
        }
        self.assertEquals(buckets, expected_output)

    def test_value_aggregation4(self):
        """Test 2 tier aggregation with probes."""
        keys = [ValueKeyAggregator(key='status'), ValueKeyAggregator(key='asn')]
        buckets = aggregate(self.probes, keys)
        expected_output = {
            'STATUS: Connected': {
                'ASN: 338': [self.Probe(id=8, country='DE', asn=338, status='Connected')],
                'ASN: 348': [self.Probe(id=9, country='DK', asn=348, status='Connected')],
                'ASN: 333': [self.Probe(id=1, country='GR', asn=333, status='Connected')],
                'ASN: 334': [self.Probe(id=2, country='NL', asn=334, status='Connected')],
                'ASN: 335': [self.Probe(id=3, country='SE', asn=335, status='Connected'), self.Probe(id=7, country='IN', asn=335, status='Connected')]
            },
            'STATUS: DisConnected': {
                'ASN: 336': [self.Probe(id=4, country='SE', asn=336, status='DisConnected')],
                'ASN: 337': [self.Probe(id=5, country='SE', asn=337, status='DisConnected')],
                'ASN: 340': [self.Probe(id=11, country='DE', asn=340, status='DisConnected')],
                'ASN: 335': [self.Probe(id=6, country='SE', asn=335, status='DisConnected')]
            },
            'STATUS: NeverConnected': {
                'ASN: 338': [self.Probe(id=10, country='DE', asn=338, status='NeverConnected')]
            }
        }
        self.assertEquals(buckets, expected_output)

    def test_range_aggregation(self):
        """Test 1 tier range aggregation"""
        keys = [RangeKeyAggregator(ranges=[10, 20, 30, 40, 50], key='rtt')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'RTT: 10-20': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8'), self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')],
            'RTT: 20-30': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8'), self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')],
            'RTT: 30-40': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8'), self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8'), self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8'), self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')],
            'RTT: 40-50': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')],
            'RTT: < 10': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8'), self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')]
        }
        self.maxDiff = None
        import pprint
        pp = pprint.PrettyPrinter()
        pp.pprint(buckets)
        self.assertEquals(buckets, expected_output)

    def test_range_aggregation1(self):
        """Test 1 tier range aggregation (1)"""
        keys = [RangeKeyAggregator(ranges=[10, 20, 30, 35, 40, 50, 60], key='rtt')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'RTT: 10-20': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8'), self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')],
            'RTT: 20-30': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8'), self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')],
            'RTT: 30-35': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8'), self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8'), self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')],
            'RTT: 35-40': [self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')],
            'RTT: 40-50': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')],
            'RTT: < 10': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8'), self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')]
        }
        self.assertEquals(buckets, expected_output)

    def test_range_aggregation2(self):
        """Test 1 tier range aggregation (2)"""
        keys = [RangeKeyAggregator(ranges=[1, 5, 10, 20, 30, 35], key='rtt')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'RTT: 1-5': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')],
            'RTT: 10-20': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8'), self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')],
            'RTT: 20-30': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8'), self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')],
            'RTT: 30-35': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8'), self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8'), self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')],
            'RTT: 5-10': [self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')],
            'RTT: > 35': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8'), self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')]}
        self.assertEquals(buckets, expected_output)

    def test_mixed_aggregation(self):
        """Test value and range aggregation together"""
        keys = [RangeKeyAggregator(ranges=[1, 5, 10, 20, 30, 35], key='rtt'), ValueKeyAggregator(key='probe.country')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'RTT: 1-5': {
                'COUNTRY: GR': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')]},
            'RTT: 10-20': {
                'COUNTRY: SE': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8'), self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')]
            },
            'RTT: 20-30': {
                'COUNTRY: DE': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8'), self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')]
            },
            'RTT: 30-35': {
                'COUNTRY: IN': [self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')],
                'COUNTRY: NL': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8')],
                'COUNTRY: SE': [self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8')]},
            'RTT: 5-10': {'COUNTRY: SE': [self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')]},
            'RTT: > 35': {
                'COUNTRY: DE': [self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')],
                'COUNTRY: DK': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')]
            }
        }
        self.assertEquals(buckets, expected_output)

    def test_mixed_aggregation1(self):
        """Test value and range aggregation together (1)"""
        keys = [RangeKeyAggregator(ranges=[10, 200], key='rtt'), ValueKeyAggregator(key='probe.country'), ValueKeyAggregator(key='probe.status')]
        buckets = aggregate(self.results, keys)
        expected_output = {
            'RTT: 10-200': {
                'COUNTRY: DE': {
                    'STATUS: Connected': [self.Result(id=8, probe=self.Probe(id=8, country='DE', asn=338, status='Connected'), rtt=28, source='127.0.0.1', prefix='192/8')],
                    'STATUS: DisConnected': [self.Result(id=11, probe=self.Probe(id=11, country='DE', asn=340, status='DisConnected'), rtt=40, source='127.0.0.1', prefix='193/8')],
                    'STATUS: NeverConnected': [self.Result(id=10, probe=self.Probe(id=10, country='DE', asn=338, status='NeverConnected'), rtt=28, source='127.0.0.1', prefix='194/8')]
                },
                'COUNTRY: DK': {'STATUS: Connected': [self.Result(id=9, probe=self.Probe(id=9, country='DK', asn=348, status='Connected'), rtt=48, source='127.0.0.1', prefix='195/8')]},
                'COUNTRY: IN': {'STATUS: Connected': [self.Result(id=7, probe=self.Probe(id=7, country='IN', asn=335, status='Connected'), rtt=35, source='127.0.0.1', prefix='192/8')]},
                'COUNTRY: NL': {'STATUS: Connected': [self.Result(id=2, probe=self.Probe(id=2, country='NL', asn=334, status='Connected'), rtt=34.0, source='127.0.0.1', prefix='192/8')]},
                'COUNTRY: SE': {
                    'STATUS: Connected': [self.Result(id=3, probe=self.Probe(id=3, country='SE', asn=335, status='Connected'), rtt=35.0, source='127.0.0.1', prefix='193/8')],
                    'STATUS: DisConnected': [self.Result(id=5, probe=self.Probe(id=5, country='SE', asn=337, status='DisConnected'), rtt=17, source='127.0.0.1', prefix='192/8'), self.Result(id=6, probe=self.Probe(id=6, country='SE', asn=335, status='DisConnected'), rtt=15, source='127.0.0.1', prefix='195/8')]
                }
            },
            'RTT: < 10': {
                'COUNTRY: GR': {'STATUS: Connected': [self.Result(id=1, probe=self.Probe(id=1, country='GR', asn=333, status='Connected'), rtt=3, source='127.0.0.1', prefix='192/8')]},
                'COUNTRY: SE': {'STATUS: DisConnected': [self.Result(id=4, probe=self.Probe(id=4, country='SE', asn=336, status='DisConnected'), rtt=6, source='127.0.0.1', prefix='193/8')]}
            }
        }
        self.assertEquals(buckets, expected_output)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function

import random
import requests

from ..cache import cache
from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from .base import Command as BaseCommand


class Command(BaseCommand):

    DESCRIPTION = "https://xkcd.com/806/"
    HEADERS = {
        "Accept": "application/vnd.github.v3+json",
        "User-Agent": "RIPE Atlas Tools (Magellan)",
    }
    URLS = {
        "root": "https://api.github.com",
        "statistics": [
            "/repos/RIPE-NCC/ripe.atlas.sagan/stats/contributors",
            "/repos/RIPE-NCC/ripe-atlas-cousteau/stats/contributors",
            "/repos/RIPE-NCC/ripe-atlas-tools/stats/contributors",
        ],
        "users": "/users"
    }

    SPACING = (
        61, 61, 61, 53, 7, 53, 6, 52, 5, 52, 49, 48, 47, 46, 47, 47, 43, 41, 38,
        39, 42, 46
    )
    BOAT = (
        "\n{}|\n{}|\n{}|\n{}|{}|\n{}|{}---\n{}---{}'-'\n{}'-'  ____|_____\n{}__"
        "__|__/    |    /\n{}/    | /     |   /\n{}/     |(      |  (\n{}(     "
        " | \     |   \\\n{}\     |  \____|____\   /|\n{}/\____|___`---.----` ."
        "' |\n{}.-'/      |  \    |__.--'    \\\n{}.'/ (       |   \   |.      "
        "    \\\n{}_ /_/   \      |    \  | `.         \\\n{}`-.'    \.--._|.--"
        "-`  |   `-._______\\\n{}``-.-------'-------'------------/\n{}`'.______"
        "_________________.'\n"
    ).format(*[" " * _ for _ in SPACING])

    WATER = ("~" * 80)

    def __init__(self, *args, **kwargs):
        BaseCommand.__init__(self, *args, **kwargs)
        self.statistics = {}

    def run(self):

        r = (
            "\nThanks for using RIPE Atlas!\n\nThis toolkit "
            "(Magellan) is a group effort, spearheaded by the team at the "
            "RIPE\nNCC, but supported by members of the community from all "
            "over.  If you're\ncurious about who we are and what sorts of "
            "stuff we work on, here's a break\ndown of our contributions to "
            "date.\n\nName                     Changes  URL\n{}\n"
        ).format("-" * 79)

        for contributor in self.get_contributors():
            r += u"{name:20}  {changes:10}  {url}\n".format(**contributor)

        print(u"{}{}{}\n".format(
            r,
            colourise(colourise(self.BOAT, "black"), "bold"),
            colourise(self.WATER, "blue")
        ))

    def get_contributors(self):

        cache_key = "github:statistics"

        self.statistics = cache.get(cache_key, {})
        if not self.statistics:
            for url in self.URLS["statistics"]:
                self._update_statistics_from_url(url)
            cache.set(cache_key, self.statistics, 60 * 10)

        r = []
        for k, v in self.statistics.items():
            r.append({
                "name": sanitise(k),
                "changes": v["changes"],
                "url": v["url"]
            })

        random.shuffle(r)

        return r

    def _update_statistics_from_url(self, url):

        response = requests.get(
            "{}{}".format(self.URLS["root"], url), headers=self.HEADERS)

        contributors = response.json()

        # Sometimes, GitHub just returns nothing
        if not contributors:
            return self._update_statistics_from_url(url)

        for contributor in response.json():

            user = self.get_user(contributor["author"]["login"])
            name = user["name"] or contributor["author"]["login"]

            if name not in self.statistics:
                self.statistics[name] = {
                    "changes": 0,
                    "url": contributor["author"]["html_url"]
                }

            for week in contributor["weeks"]:
                self.statistics[name]["changes"] += week["a"] + week["d"]

    def get_user(self, username):

        cache_key = "github-user:{}".format(username)

        user = cache.get(cache_key)
        if user:
            return user

        cache.set(
            cache_key,
            requests.get(
                "{}{}/{}".format(
                    self.URLS["root"],
                    self.URLS["users"],
                    username
                ),
                headers=self.HEADERS
            ).json(),
            60 * 60 * 24 * 365
        )

        return self.get_user(username)



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import requests

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

from ripe.atlas.tools.commands.probe_search import Command
from ripe.atlas.tools.exceptions import RipeAtlasToolsException
from ripe.atlas.cousteau import Probe
from ripe.atlas.tools.aggregators import ValueKeyAggregator

from ..base import capture_sys_output


COMMAND_MODULE = "ripe.atlas.tools.commands.probe_search"


class FakeGen(object):
    def __init__(self,):
        self.probes = [
            Probe(id=1, meta_data={
                "country_code": "GR", "asn_v4": 3333, "prefix_v4": "193.0/22"}),
            Probe(id=2, meta_data={
                "country_code": "DE", "asn_v4": 3333, "prefix_v4": "193.0/22"}),
            Probe(id=3, meta_data={
                "country_code": "DE", "asn_v4": 3332, "prefix_v4": "193.0/22"}),
            Probe(id=4, meta_data={
                "country_code": "NL", "asn_v4": 3333, "prefix_v4": "193.0/22"}),
            Probe(id=5, meta_data={
                "country_code": "GR", "asn_v4": 3333, "prefix_v4": "193.0/22"}),
        ]
        self.total_count = 4

    def __iter__(self):
        return self

    # Python 3 compatibility
    def __next__(self):
        return self.next()

    def next(self):
        if not self.probes:
            raise StopIteration()
        else:
            return self.probes.pop(0)


class TestProbesCommand(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None

    def test_with_empty_args(self):
        """User passes no args, should fail with RipeAtlasToolsException"""
        with self.assertRaises(RipeAtlasToolsException):
            cmd = Command()
            cmd.init_args([])
            cmd.run()

    def test_with_random_args(self):
        """User passes random args, should fail with SystemExit"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["blaaaaaaa"])
                cmd.run()

    def test_arg_with_no_value(self):
        """User passed not boolean arg but no value"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--asn"])
                cmd.run()

    def test_arg_with_wrong_type(self):
        """User passed arg with wrong type. e.g string for asn"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--asn", "blaaaaa"])
                cmd.run()
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--asnv4", "blaaaaa"])
                cmd.run()
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--asnv6", "blaaaaa"])
                cmd.run()
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--limit", "blaaaaa"])
                cmd.run()
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--radius", "blaaaaa"])
                cmd.run()
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--tag", "inv@lid"])
                cmd.run()

    def test_location_google_breaks(self):
        """User passed location arg but google api gave error"""
        caught_exceptions = [
            requests.ConnectionError, requests.HTTPError, requests.Timeout]
        with mock.patch('requests.get') as mock_get:
            for exception in caught_exceptions:
                mock_get.side_effect = exception
                with capture_sys_output():
                    with self.assertRaises(RipeAtlasToolsException):
                        cmd = Command()
                        cmd.init_args(["--location", "blaaaa"])
                        cmd.run()
            mock_get.side_effect = Exception()
            with self.assertRaises(Exception):
                cmd = Command()
                cmd.init_args(["--location", "blaaaa"])
                cmd.run()

    def test_location_google_wrong_output(self):
        """User passed location arg but google api gave not expected format"""
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value = requests.Response()
            with mock.patch('requests.Response.json') as mock_json:
                mock_json.return_value = {"blaaa": "bla"}
                with self.assertRaises(RipeAtlasToolsException):
                    cmd = Command()
                    cmd.init_args(["--location", "blaaaa"])
                    cmd.run()

    def test_location_arg(self):
        """User passed location arg"""
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value = requests.Response()
            with mock.patch('requests.Response.json') as mock_json:
                mock_json.return_value = {"results": [
                    {"geometry": {"location": {"lat": 1, "lng": 2}}}]}
                cmd = Command()
                cmd.init_args(["--location", "blaaaa"])
                self.assertEquals(cmd.build_request_args(), {'radius': '1,2:15'})

    def test_location_arg_with_radius(self):
        """User passed location arg"""
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value = requests.Response()
            with mock.patch('requests.Response.json') as mock_json:
                mock_json.return_value = {"results": [
                    {"geometry": {"location": {"lat": 1, "lng": 2}}}
                ]}
                cmd = Command()
                cmd.init_args(["--location", "blaaaa", "--radius", "4"])
                self.assertEquals(
                    cmd.build_request_args(),
                    {"radius": "1,2:4"}
                )

    def test_asn_args(self):
        """User passed asn arg together with asnv4 or asnv6"""
        with self.assertRaises(RipeAtlasToolsException):
            cmd = Command()
            cmd.init_args(["--asn", "3333", "--asnv4", "3333"])
            cmd.run()

        with self.assertRaises(RipeAtlasToolsException):
            cmd = Command()
            cmd.init_args(["--asn", "3333", "--asnv6", "3333"])
            cmd.run()

    def test_prefix_args(self):
        """User passed prefix arg together with prefixv4 or prefixv6"""
        with self.assertRaises(RipeAtlasToolsException):
            cmd = Command()
            cmd.init_args([
                "--prefix", "193.0.0.0/21",
                "--prefixv4", "193.0.0.0/21"
            ])
            cmd.run()

        with self.assertRaises(RipeAtlasToolsException):
            cmd = Command()
            cmd.init_args([
                "--prefix", "2001:67c:2e8::/48",
                "--prefixv6", "2001:67c:2e8::/48"
            ])
            cmd.run()

    def test_all_args(self):
        """User passed all arguments"""
        cmd = Command()
        cmd.init_args(["--all"])
        self.assertEquals(cmd.build_request_args(), {})

    def test_center_arg_wrong_value(self):
        """User passed center arg with wrong value"""
        with self.assertRaises(RipeAtlasToolsException):
            cmd = Command()
            cmd.init_args(["--center", "blaaaa"])
            cmd.run()

    def test_center_arg(self):
        """User passed center arg"""
        cmd = Command()
        cmd.init_args(["--center", "1,2"])
        self.assertEquals(
            cmd.build_request_args(),
            {"radius": "1,2:15"}
        )

    def test_center_arg_with_radius(self):
        """User passed center and radius arg"""
        cmd = Command()
        cmd.init_args(["--center", "1,2", "--radius", "4"])
        self.assertEquals(cmd.build_request_args(), {"radius": "1,2:4"})

    def test_country_arg(self):
        """User passed country code arg"""
        cmd = Command()
        cmd.init_args(["--country", "GR"])
        self.assertEquals(cmd.build_request_args(), {"country_code": "GR"})

    def test_country_arg_with_radius(self):
        """User passed country code arg together with radius"""
        cmd = Command()
        cmd.init_args(["--country", "GR", "--radius", "4"])
        self.assertEquals(cmd.build_request_args(), {"country_code": "GR"})

    def test_status_arg(self):
        """User passed valid status arg."""
        for status in range(0, 3):
            cmd = Command()
            cmd.init_args(["--status", str(status)])
            self.assertEquals(cmd.build_request_args(), {"status": status})

    def test_status_arg_wrong_value(self):
        """User passed status arg with wrong value"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--status", "4"])
                cmd.run()

    def test_sane_tags(self):
        """Sane tags"""
        cmd = Command()
        cmd.init_args(["--tag", "native-ipv6"])
        self.assertEquals(
            cmd.build_request_args(),
            {"tags": "native-ipv6"}
        )

        cmd = Command()
        cmd.init_args(["--tag", "native-ipv6", "--tag", "system-ipv4-works"])
        self.assertEquals(
            cmd.build_request_args(),
            {"tags": "native-ipv6,system-ipv4-works"}
        )

    def test_sane_args1(self):
        """User passed several arguments (1)"""
        cmd = Command()
        cmd.init_args([
            "--center", "1,2",
            "--radius", "4",
            "--asnv4", "3333",
            "--prefix", "193.0.0.0/21"
        ])
        self.assertEquals(
            cmd.build_request_args(),
            {'asn_v4': 3333, 'prefix': '193.0.0.0/21', 'radius': '1,2:4'}
        )

    def test_sane_args2(self):
        """User passed several arguments (2)"""

        cmd = Command()
        cmd.init_args([
            "--location", "Amsterdam",
            "--asn", "3333",
            "--prefixv4", "193.0.0.0/21"
        ])

        path = '{}.Command.location2degrees'.format(COMMAND_MODULE)
        with mock.patch(path) as mock_get:
            mock_get.return_value = (1, 2)
            self.assertEquals(cmd.build_request_args(), {
                'asn': 3333,
                'prefix_v4': '193.0.0.0/21',
                'radius': '1,2:15'
            })

    def test_sane_args3(self):
        """User passed several arguments (3)"""

        cmd = Command()
        cmd.init_args([
            "--center", "1,2",
            "--asnv6", "3333",
            "--prefixv6", "2001:67c:2e8::/48"
        ])
        self.assertEquals(cmd.build_request_args(), {
            'asn_v6': 3333,
            'prefix_v6': '2001:67c:2e8::/48',
            'radius': '1,2:15'
        })

    def test_render_ids_only(self):
        """User passed ids_only arg, testing rendiring"""

        cmd = Command()
        cmd.init_args([
            "--ids-only", "--country", "GR"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                self.assertEquals(stdout.getvalue(), "1\n2\n3\n4\n5\n")

    def test_render_ids_only_with_limit(self):
        """User passed ids_only arg together with limit, testing rendering"""
        cmd = Command()
        cmd.init_args([
            "--ids-only",
            "--country", "GR",
            "--limit", "2"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                self.assertEquals(stdout.getvalue(), "1\n2\n")

    def test_render_ids_only_with_aggr(self):
        """
        User passed ids_only arg together with aggregate, testing rendering
        """
        cmd = Command()
        cmd.init_args([
            "--ids-only",
            "--country", "GR",
            "--aggregate-by", "country"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                self.assertEquals(stdout.getvalue(), "1\n2\n3\n4\n5\n")

    def test_get_aggregators(self):
        """User passed --aggregate-by args"""
        cmd = Command()
        cmd.init_args([
            "--aggregate-by", "asn_v4",
            "--aggregate-by", "country",
            "--aggregate-by", "prefix_v4"
        ])
        expected_output = [
            ValueKeyAggregator(key="asn_v4"),
            ValueKeyAggregator(key="country_code"),
            ValueKeyAggregator(key="prefix_v4")
        ]
        cmd.set_aggregators()
        for index, v in enumerate(cmd.aggregators):
            self.assertTrue(isinstance(v, ValueKeyAggregator))
            self.assertEquals(
                v.aggregation_keys,
                expected_output[index].aggregation_keys
            )

    def test_render_without_aggregation(self):
        """Tests rendering of results without aggregation"""
        cmd = Command()
        cmd.init_args([
            "--country", "GR"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                expected_output = [
                    "",
                    "Filters:",
                    "  Country: GR",
                    "",
                    "ID    Asn_v4 Asn_v6 Country Status         ",
                    "===========================================",
                    str(b"1     3333            gr    None           "),
                    str(b"2     3333            de    None           "),
                    str(b"3     3332            de    None           "),
                    str(b"4     3333            nl    None           "),
                    str(b"5     3333            gr    None           "),
                    "===========================================",
                    "                Showing 4 of 4 total probes",
                    "",
                ]
                self.assertEquals(set(stdout.getvalue().split("\n")), set(expected_output))

    def test_render_without_aggregation_with_limit(self):
        """Tests rendering of results without aggregation but with limit"""
        cmd = Command()
        cmd.init_args([
            "--country", "GR",
            "--limit", "2"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = 'ripe.atlas.tools.commands.probe_search.ProbeRequest'
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                self.maxDiff = None
                expected_output = [
                    "",
                    "Filters:",
                    "  Country: GR",
                    "",
                    "ID    Asn_v4 Asn_v6 Country Status         ",
                    "===========================================",
                    str(b"1     3333            gr    None           "),
                    str(b"2     3333            de    None           "),
                    "===========================================",
                    "                Showing 2 of 4 total probes",
                    ""
                ]
                self.assertEquals(set(stdout.getvalue().split("\n")), set(expected_output))

    def test_render_with_aggregation(self):
        """Tests rendering of results with aggregation"""
        cmd = Command()
        cmd.init_args([
            "--country", "GR",
            "--aggregate-by", "country",
            "--aggregate-by", "asn_v4",
            "--aggregate-by", "prefix_v4"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                expected_blob = [
                    "",
                    "Filters:",
                    "  Country: GR",
                    "",
                    "   ID    Asn_v4 Asn_v6 Country Status         ",
                    "==============================================",
                    "Country: DE",
                    " ASN_V4: 3332",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   3     3332            de    None           "),
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   2     3333            de    None           "),
                    "",
                    "Country: GR",
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   1     3333            gr    None           "),
                    str(b"   5     3333            gr    None           "),
                    "",
                    "Country: NL",
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   4     3333            nl    None           "),
                    "==============================================",
                    "                   Showing 4 of 4 total probes",
                    ""
                ]
                self.maxDiff = None
                out = stdout.getvalue()
                expected_set = set(expected_blob)
                returned_set = set(out.split("\n"))
                self.assertEquals(returned_set, expected_set)

    def test_render_with_aggregation_with_limit(self):
        """Tests rendering of results with aggregation with limit"""
        cmd = Command()
        cmd.init_args([
            "--country", "GR",
            "--aggregate-by", "country",
            "--aggregate-by", "asn_v4",
            "--aggregate-by", "prefix_v4",
            "--limit", "1"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                expected_output = [
                    "",
                    "Filters:",
                    "  Country: GR",
                    "",
                    "   ID    Asn_v4 Asn_v6 Country Status         ",
                    "==============================================",
                    "Country: GR",
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   1     3333            gr    None           "),
                    "==============================================",
                    "                   Showing 1 of 4 total probes",
                    "",
                ]
                expected_set = set(expected_output)
                returned_set = set(stdout.getvalue().split("\n"))
                self.assertEquals(returned_set, expected_set)

    def test_render_with_aggregation_with_max_per_aggr(self):
        """
        Tests rendering of results with aggregation with max per aggr option
        """
        cmd = Command()
        cmd.init_args([
            "--country", "GR",
            "--aggregate-by", "country",
            "--aggregate-by", "asn_v4",
            "--aggregate-by", "prefix_v4",
            "--max-per-aggregation", "1"
        ])

        with capture_sys_output() as (stdout, stderr):
            path = '{}.ProbeRequest'.format(COMMAND_MODULE)
            with mock.patch(path) as mock_get:
                mock_get.return_value = FakeGen()
                cmd.run()
                expected_output = [
                    "",
                    "Filters:",
                    "  Country: GR",
                    "",
                    "   ID    Asn_v4 Asn_v6 Country Status         ",
                    "==============================================",
                    "Country: DE",
                    " ASN_V4: 3332",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   3     3332            de    None           "),
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   2     3333            de    None           "),
                    "",
                    "Country: GR",
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   1     3333            gr    None           "),
                    "",
                    "Country: NL",
                    " ASN_V4: 3333",
                    "  PREFIX_V4: 193.0/22",
                    str(b"   4     3333            nl    None           "),
                    "==============================================",
                    "                   Showing 4 of 4 total probes",
                    "",
                ]
                expected_set = set(expected_output)
                returned_set = set(stdout.getvalue().split("\n"))
                self.assertEquals(returned_set, expected_set)



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import datetime
import os
import sys
import unittest

try:
    from cStringIO import StringIO
except ImportError:  # Python 3
    from io import StringIO

from ripe.atlas.tools.helpers.validators import ArgumentType


class TestArgumentTypeHelper(unittest.TestCase):

    def test_path(self):

        self.assertEqual("/tmp", ArgumentType.path("/tmp"))

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.path("/not/a/real/place")

    def test_country_code(self):

        self.assertEqual("CA", ArgumentType.country_code("CA"))
        self.assertEqual("CA", ArgumentType.country_code("ca"))
        self.assertEqual("CA", ArgumentType.country_code("Ca"))
        self.assertEqual("CA", ArgumentType.country_code("cA"))

        for value in ("CAN", "Canada", "can", "This isn't even a country"):
            with self.assertRaises(argparse.ArgumentTypeError):
                ArgumentType.country_code(value)

    def test_comma_separated_integers(self):

        self.assertEqual(
            [1, 2, 3], ArgumentType.comma_separated_integers()("1,2,3"))

        self.assertEqual(
            [1, 2, 3], ArgumentType.comma_separated_integers()("1, 2, 3"))

        self.assertEqual([1], ArgumentType.comma_separated_integers()("1"))

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.comma_separated_integers()("1,2,3,pizza!")
        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.comma_separated_integers(minimum=5)("4,5,6,7")
        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.comma_separated_integers(maximum=5)("1,2,3,4,6")

    def test_datetime(self):

        d = datetime.datetime(2015, 12, 1)

        self.assertEqual(d, ArgumentType.datetime("2015-12-1"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1T00"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1T00:00"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1T00:00:00"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1 00"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1 00:00"))
        self.assertEqual(d, ArgumentType.datetime("2015-12-1 00:00:00"))

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.datetime("yesterday")

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.datetime("Definitely not a date, or even a time")

    def test_integer_range(self):

        self.assertEqual(1, ArgumentType.integer_range(1, 10)("1"))
        self.assertEqual(10, ArgumentType.integer_range(1, 10)("10"))
        self.assertEqual(1, ArgumentType.integer_range(-1, 1)("1"))
        self.assertEqual(-1, ArgumentType.integer_range(-1, 1)("-1"))

        for value in ("0", "11", "-1"):
            with self.assertRaises(argparse.ArgumentTypeError):
                ArgumentType.integer_range(1, 10)(value)

    def test_ip_or_domain(self):

        passable_hosts = (
            "localhost", "ripe.net", "www.ripe.net", "1.2.3.4",
            "2001:67c:2e8:22::c100:68b"
        )
        for host in passable_hosts:
            self.assertEqual(host, ArgumentType.ip_or_domain(host))

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.ip_or_domain("Definitely not a host")

    def test_comma_separated_integers_or_file(self):

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.comma_separated_integers_or_file("/dev/null/fail")

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.comma_separated_integers_or_file("not,a,number")

        with self.assertRaises(argparse.ArgumentTypeError):
            ArgumentType.comma_separated_integers_or_file("1, 2, 3")

        old = sys.stdin
        sys.stdin = StringIO("1\n2\n")
        self.assertEqual(
            ArgumentType.comma_separated_integers_or_file("-"),
            [1, 2]
        )
        sys.stdin = old

        in_file = "/tmp/__test_file__"
        with open(in_file, "w") as f:
            f.write("1\n2\n3\n")
        self.assertEqual(
            ArgumentType.comma_separated_integers_or_file(in_file),
            [1, 2, 3]
        )
        os.unlink(in_file)

    def test_measurement_alias(self):

        tests = ["", "\\invalid", "+invalid",
                 ":invalid", "12345"]
        for test in tests:
            with self.assertRaises(argparse.ArgumentTypeError):
                ArgumentType.alias_is_valid(test)

        tests = ["valid", "123valid", "valid123", "_valid",
                 "valid_", "-valid", "valid-", ".valid"]

        for test in tests:
            self.assertEqual(
                ArgumentType.alias_is_valid(test),
                test
            )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from ..exceptions import RipeAtlasToolsException
from ..helpers.validators import ArgumentType
from ..settings import AliasesDB, aliases
from .base import Command as BaseCommand


class Command(BaseCommand):

    NAME = "alias"

    EDITOR = os.environ.get("EDITOR", "/usr/bin/vim")
    DESCRIPTION = "Manage measurements' and probes' aliases"
    EXTRA_DESCRIPTION = (
        "As an alternative to this command, you can just create/edit {}"
        .format(AliasesDB.USER_RC)
    )

    def add_arguments(self):
        subparsers = self.parser.add_subparsers(
            title="action",
            dest="action",
            help="Action to be performed on aliases. "
                 "Run 'ripe-atlas alias <action> --help' for more details."
        )

        add_parser = subparsers.add_parser(
            "add",
            help="Add/modify an alias."
        )
        add_parser.add_argument(
            "type",
            action="store",
            choices=["measurement", "probe"],
            help="Type of target object.",
        )
        add_parser.add_argument(
            "target",
            action="store",
            type=int,
            help="Target's ID.",
        )
        add_parser.add_argument(
            "alias",
            action="store",
            type=ArgumentType.alias_is_valid,
            help="Alias name.",
        )

        del_parser = subparsers.add_parser(
            "del",
            help="Remove an alias."
        )
        del_parser.add_argument(
            "type",
            action="store",
            choices=["measurement", "probe"],
            help="Type of target object.",
        )
        del_parser.add_argument(
            "alias",
            action="store",
            type=ArgumentType.alias_is_valid,
            help="Alias name.",
        )

        show_parser = subparsers.add_parser(
            "show",
            help="Show target's ID."
        )
        show_parser.add_argument(
            "type",
            action="store",
            choices=["measurement", "probe"],
            help="Type of target object.",
        )
        show_parser.add_argument(
            "alias",
            action="store",
            type=ArgumentType.alias_is_valid,
            help="Alias name.",
        )

        list_parser = subparsers.add_parser(
            "list",
            help="List configured aliases."
        )
        list_parser.add_argument(
            "type",
            action="store",
            choices=["measurement", "probe"],
            help="Type of target object.",
        )

        subparsers.add_parser(
            "editor",
            help="Invoke {0} to edit the configuration directly".format(
                self.EDITOR)
        )

    def run(self):

        if not self.arguments.action:
            raise RipeAtlasToolsException(
                "Action not given. Use --help for more information.")

        if self.arguments.action == "editor":
            os.system("{0} {1}".format(self.EDITOR, AliasesDB.USER_RC))
            return self.ok(
                "Aliases file writen to {}".format(AliasesDB.USER_RC))

        else:
            alias_type = self.arguments.type

            if self.arguments.action == "add":
                alias_name = self.arguments.alias
                target_id = self.arguments.target

                try:
                    ArgumentType.alias_is_valid(alias_name)
                except Exception as e:
                    raise RipeAtlasToolsException(str(e))

                aliases[alias_type][alias_name] = target_id
                AliasesDB.write(aliases)

            elif self.arguments.action == "del":
                alias_name = self.arguments.alias
                del aliases[alias_type][alias_name]
                AliasesDB.write(aliases)

            elif self.arguments.action == "show":
                alias_name = self.arguments.alias
                if alias_name in aliases[alias_type]:
                    self.ok(
                        "'{}' is an alias for {}".format(
                            alias_name,
                            aliases[alias_type][alias_name]
                        )
                    )
                else:
                    self.not_ok(
                        "'{}' alias does not exist".format(
                            alias_name
                        )
                    )

            elif self.arguments.action == "list":
                res = "{} aliases:\n\n".format(alias_type.capitalize())
                for alias_name in sorted(aliases[alias_type]):
                    res += "- {}: {}\n".format(
                        alias_name, aliases[alias_type][alias_name]
                    )
                self.ok(res)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ...exceptions import RipeAtlasToolsException
from ..base import Factory as BaseFactory
from .ping import PingMeasureCommand
from .traceroute import TracerouteMeasureCommand
from .dns import DnsMeasureCommand
from .sslcert import SslcertMeasureCommand
from .http import HttpMeasureCommand
from .ntp import NtpMeasureCommand


class Factory(BaseFactory):

    TYPES = {
        "ping": PingMeasureCommand,
        "traceroute": TracerouteMeasureCommand,
        "dns": DnsMeasureCommand,
        "sslcert": SslcertMeasureCommand,
        "http": HttpMeasureCommand,
        "ntp": NtpMeasureCommand,
    }
    DESCRIPTION = "Create a measurement and wait for the results"

    def __init__(self, sys_args):

        self.build_class = None
        self.sys_args = sys_args
        if len(self.sys_args) >= 2:
            self.build_class = self.TYPES.get(self.sys_args[1].lower())

        if not self.build_class:
            self.raise_log()

    def raise_log(self):
        """Depending on the input raise with different log message."""
        # cases: 1) ripe-atlas measure 2) ripe-atlas measure --help/-h
        if (
            len(self.sys_args) == 1 or
            (len(self.sys_args) == 2 and self.sys_args[1] in ("--help", "-h"))
        ):
            log = (
                "Usage: ripe-atlas measure <type> [arguments]\n\n"
                "Types:\n"
            )
            for type_ in sorted(self.TYPES):
                log += "\t{}\n".format(type_)
            log += (
                "\nFor extended options for a specific measurement type, "
                "try ripe-atlas measure <type> --help."
            )
        # cases: ripe-atlas measure bla
        else:
            log = (
                "The measurement type you requested is invalid.  "
                "Please choose one of {}."
            ).format(", ".join(self.TYPES.keys()))
        raise RipeAtlasToolsException(log)

    def create(self, *args, **kwargs):
        return self.build_class(*args, **kwargs)



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .ping import TestPingRenderer
from .http import TestHttpRenderer
from .aggregate_ping import TestAggregatePing
from .ssl_consistency import TestSSLConsistency
from .raw import TestRawRenderer
from .dns_compact import TestDnsCompact

__all__ = [
    TestPingRenderer,
    TestHttpRenderer,
    TestAggregatePing,
    TestSSLConsistency,
    TestDnsCompact,
]



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tzlocal import get_localzone
from datetime import datetime
from .base import Renderer as BaseRenderer
from ..helpers.colours import colourise


class Renderer(BaseRenderer):
    RENDERS = [BaseRenderer.TYPE_NTP]

    def on_result(self, result):
        created = result.created.astimezone(get_localzone())
        probe_id = result.probe_id
        r = '\n\nProbe #{0}\n{1}\n'.format(probe_id, '=' * 79)
        res = self.get_formatted_response(probe_id, created, result)
        if res:
            return r + res
        else:
            return colourise(r + 'No results\n', 'red')

    def get_formatted_response(self, probe_id, created, result):
        leap = result.leap_second_indicator
        stratum = result.stratum
        v = result.version
        mode = result.mode
        # just here for completeness, flake8 doesn't like when it's not used
        # end_time = result.end_time
        poll = result.poll
        precision = result.precision
        refid = result.reference_id
        ref_time = result.reference_time
        root_delay = result.root_delay
        root_disp = result.root_dispersion

        if not leap and not stratum and not v and not mode:
            return

        if mode != 'server':
            print('invalid mode: %s' % mode)

        r = '[NTP] %s -> %s (%s)\n' % (result.source_address,
                                       result.destination_name,
                                       result.destination_address)
        r += '\tversion: %s, stratum: %s/16\n' % (colourise(v, 'bold'),
                                                  colourise(stratum, 'bold'))
        r += '\trefid: %s\n' % colourise(refid, 'bold')
        r += '\tleap: %s, poll: %s, precision: %s\n' % (leap, poll, precision)
        r += '\troot_delay: %s, root_disp: %s\n' % (root_delay, root_disp)
        r += '\tref-time: %s\n\n' % datetime.fromtimestamp(ref_time)

        try:
            for idx, pkt in enumerate(result.packets):
                r += '\t[%s] %s\n' % (idx, str(pkt))
                r += '\t\ttrans: %s -> recv: %s\n' % (pkt.transmitted_time,
                                                      pkt.received_time)
                r += '\t\torigin: %s -> final: %s\n\n' % (pkt.origin_time,
                                                          pkt.final_time)
        except Exception as ex:
            print('Got exception when reading packet: %s' % ex)
            print('Raw: %s' % pkt.raw_data)

        return r



# coding=utf-8

# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import json
import os
import sys
import tempfile
import unittest

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

from ripe.atlas.cousteau import Probe

from ripe.atlas.tools.commands.report import Command
from ripe.atlas.tools.exceptions import RipeAtlasToolsException
from ripe.atlas.tools.renderers import Renderer
from ripe.atlas.tools.settings import AliasesDB
from ripe.atlas.tools.version import __version__
from ..base import capture_sys_output, StringIO


class TestReportCommand(unittest.TestCase):

    mocked_results = [
        {'af': 4, 'prb_id': 1216, 'result': [{'rtt': 27.429}, {'rtt': 25.672}, {'rtt': 25.681}], 'ttl': 54, 'avg': 26.2606666667, 'size': 20, 'from': '109.190.83.40', ' proto': 'ICMP', 'timestamp': 1445025400, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4700, 'max': 27.429, 'step': 360, 'src_addr': '192.168.103.132', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 377, 'dst_name': 'hsi.cablecom.ch', 'min': 25.672, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 165, 'result': [{'rtt': 87.824658}, {'rtt': 87.611154}, {'rtt': 90.99957}], 'ttl': 48, 'avg': 88.811794, 'size': 20, 'from': '194.85.27.7', 'proto': 'ICMP', 'timestamp': 1445025590, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 90.99957, 'step': 360, 'src_addr': '192.168.3.8', 'rcvd': 3, 'msm_name': 'Ping', ' lts': 87, 'dst_name': 'hsi.cablecom.ch', 'min': 87.611154, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 202, 'result': [{'rtt': 40.02356}, {'rtt': 40.399112}, {'rtt': 39.29012}], 'ttl': 52, 'avg': 39.904264, 'size': 20, 'from': '178.190.51.206', 'proto': 'ICMP', 'timestamp': 1445015502, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 40.399112, 'step': 360, 'src_addr': '10.0.0.2', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 502, 'dst_name': 'hsi.cablecom.ch', 'min': 39.29012, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 2225, 'result': [{'rtt': 10.858}, {'rtt': 12.632}, {'rtt': 20.53}, {'dup': 1, 'rtt': 32.775}, {'dup': 1, 'rtt': 47.509}, {'dup': 1, 'rtt': 62.745}, {'dup': 1, 'rtt': 78.54}, {'dup': 1, 'rtt': 93.272}, {'dup': 1, 'rtt': 109.738}], 'ttl': 56, 'avg': 14.6733333333, 'size': 20, 'from': '46.126.90.165', 'proto': 'ICMP', 'timestamp': 1445025616, 'dup': 6, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4700, 'max': 20.53, 'step': 360, 'src_addr': '192.168.111.103', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 309, 'dst_name': 'hsi.cablecom.ch', 'min': 10.858, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 270, 'result': [{'rtt': 28.527366}, {'rtt': 26.585862}, {'rtt': 26.393094}], 'ttl': 51, 'avg': 27.168774, 'size': 20, 'from': '188.192.110.111', 'proto': 'ICMP', 'timestamp': 1445025513, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 28.527366, 'step': 360, 'src_addr': '192.168.178.21', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 182, 'dst_name': 'hsi.cablecom.ch', 'min': 26.393094, 'group_id': 1000192, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 579, 'result': [{'rtt': 23.201285}, {'rtt': 22.980868}, {'rtt': 22.863364}], 'ttl': 51, 'avg': 23.0151723333, 'size': 20, 'from': '195.88.195.170', 'proto': 'ICMP', 'timestamp': 1445025521, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 23.201285, 'step': 360, 'src_addr': '10.69.8.150', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 56, 'dst_name': 'hsi.cablecom.ch', 'min': 22.863364, 'group_id': 1000192, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 677, 'result': [{'rtt': 40.71476}, {'rtt': 40.258568}, {'rtt': 40.316936}], 'ttl': 54, 'avg': 40.430088, 'size': 20, 'from': '78.128.9.202', 'proto': 'ICMP', 'timestamp': 1445025298, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 40.71476, 'step': 360, 'src_addr': '10.100.0.25', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 97, 'dst_name': 'hsi.cablecom.ch', 'min': 40.258568, 'group_id': 1000192, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 879, 'result': [{'rtt': 34.319623}, {'rtt': 34.445575}, {'rtt': 34.376455}], 'ttl': 53, 'avg': 34.380551, 'size': 20, 'from': '94.254.125.2', 'proto': 'ICMP', 'timestamp': 1445025223, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 34.445575, 'step': 360, 'src_addr': '192.168.8.130', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 189, 'dst_name': 'hsi.cablecom.ch', 'min': 34.319623, 'dst_addr': '62.2.16.24'},
        {'af': 4, 'prb_id': 945, 'result': [{'rtt': 61.665036}, {'rtt': 23.833349}, {'rtt': 23.268868}], 'ttl': 56, 'avg': 36.255751, 'size': 20, 'from': '92.111.237.94', 'proto': 'ICMP', 'timestamp': 1445025494, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4720, 'max': 61.665036, 'step': 360, 'src_addr': '92.111.237.94', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 746, 'dst_name': 'hsi.cablecom.ch', 'min': 23.268868, 'dst_addr': '62.2.16.24'}
    ]
    expected_output_no_aggr = (
        "20 bytes from probe #1216  109.190.83.40   to hsi.cablecom.ch (62.2.16.24): ttl=54 times:27.429,  25.672,  25.681, \n"
        "20 bytes from probe #165   194.85.27.7     to hsi.cablecom.ch (62.2.16.24): ttl=48 times:87.825,  87.611,  91.0,   \n"
        "20 bytes from probe #202   178.190.51.206  to hsi.cablecom.ch (62.2.16.24): ttl=52 times:40.024,  40.399,  39.29,  \n"
        "20 bytes from probe #2225  46.126.90.165   to hsi.cablecom.ch (62.2.16.24): ttl=56 times:10.858,  12.632,  20.53,   32.775,  47.509,  62.745,  78.54,   93.272,  109.738,\n"
        "20 bytes from probe #270   188.192.110.111 to hsi.cablecom.ch (62.2.16.24): ttl=51 times:28.527,  26.586,  26.393, \n"
        "20 bytes from probe #579   195.88.195.170  to hsi.cablecom.ch (62.2.16.24): ttl=51 times:23.201,  22.981,  22.863, \n"
        "20 bytes from probe #677   78.128.9.202    to hsi.cablecom.ch (62.2.16.24): ttl=54 times:40.715,  40.259,  40.317, \n"
        "20 bytes from probe #879   94.254.125.2    to hsi.cablecom.ch (62.2.16.24): ttl=53 times:34.32,   34.446,  34.376, \n"
        "20 bytes from probe #945   92.111.237.94   to hsi.cablecom.ch (62.2.16.24): ttl=56 times:61.665,  23.833,  23.269, \n"
    )

    def setUp(self):
        self.cmd = Command()

    def test_with_empty_args(self):
        """User passes no args, should fail with SystemExit"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                self.cmd.init_args([])
                self.cmd.run()

    def test_with_random_args(self):
        """User passes random args, should fail with SystemExit"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["blaaaaaaa"])
                self.cmd.run()

    def test_arg_with_no_value(self):
        """User passed not boolean arg but no value"""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["--probes"])
                self.cmd.run()

    def test_arg_with_wrong_type(self):
        """User passed arg with wrong type."""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["--probes", "blaaaaa"])
                self.cmd.run()

    def test_arg_renderer_with_wrong_choice(self):
        """User passed arg renderer with unavailable type."""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["--renderer", "blaaaaa"])
                self.cmd.run()

    def test_arg_renderer_with_valid_choice(self):
        """User passed arg renderer with valid type."""
        # Mock AtlasRequest to fail run fast and test args validity.
        path = 'ripe.atlas.cousteau.AtlasRequest.get'
        with mock.patch(path) as mock_get:
            mock_get.return_value = False, {}
            for choice in Renderer.get_available():
                with self.assertRaises(RipeAtlasToolsException):
                    cmd = Command()
                    cmd.init_args(["--renderer", choice, "1"])
                    cmd.run()

    def test_arg_renderer_traceroute_aspath_with_valid_radius_arg(self):
        """User passed arg --traceroute-aspath-radius to traceroute_aspath renderer"""
        # Mock AtlasRequest to fail run fast and test args validity.
        path = 'ripe.atlas.cousteau.AtlasRequest.get'
        with mock.patch(path) as mock_get:
            mock_get.return_value = False, {}
            with self.assertRaises(RipeAtlasToolsException):
                cmd = Command()
                cmd.init_args(["--renderer", "traceroute_aspath",
                               "--traceroute-aspath-radius", "3",
                               "1"])
                cmd.run()

    def test_arg_renderer_traceroute_aspath_with_invalid_radius_arg(self):
        """User passed arg --traceroute-aspath-radius to traceroute_aspath renderer with invalid value"""
        # Mock AtlasRequest to fail run fast and test args validity.
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                cmd = Command()
                cmd.init_args(["--renderer", "traceroute_aspath",
                               "--traceroute-aspath-radius", "blaaaaa",
                               "1"])
                cmd.run()

    def test_arg_aggregate_with_valid_choice(self):
        """User passed arg aggregate with valid type."""
        # Mock AtlasRequest to fail run fast and test args validity.
        path = 'ripe.atlas.cousteau.AtlasRequest.get'
        with mock.patch(path) as mock_get:
            mock_get.return_value = False, {}
            for choice in Command.AGGREGATORS.keys():
                with self.assertRaises(RipeAtlasToolsException):
                    cmd = Command()
                    cmd.init_args(["--aggregate-by", choice, "1"])
                    cmd.run()

    def test_arg_aggregate_with_wrong_choice(self):
        """User passed arg aggregate with unavailable type."""
        with capture_sys_output():
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["--aggregate-by", "blaaaaa"])
                self.cmd.run()

    def test_arg_no_source(self):
        """User passed no measurement id and no file name."""
        with capture_sys_output(use_fake_tty=True):
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["--aggregate-by", "country"])
                self.cmd.run()

    def test_arg_from_file(self):
        """User passed a valid filename"""
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            temp_file.write(
                json.dumps(self.mocked_results).encode("utf-8")
            )
            temp_file.close()
            with capture_sys_output() as (stdout, stderr):
                self.cmd.init_args([
                    "--from-file", temp_file.name,
                ])
                self.cmd.run()
                assert self.expected_output_no_aggr == stdout.getvalue()
        finally:
            os.unlink(temp_file.name)

    def test_arg_from_stdin(self):
        """User passes results into standard input"""
        try:
            current_stdin = sys.stdin
            sys.stdin = StringIO(json.dumps(self.mocked_results))
            with capture_sys_output() as (stdout, stderr):
                self.cmd.init_args([])
                self.cmd.run()
                assert self.expected_output_no_aggr == stdout.getvalue()
        finally:
            sys.stdin = current_stdin

    def test_arg_no_valid_msm_id(self):
        """User passed non valid type of measurement id."""
        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                self.cmd.init_args(["blaaa"])
                self.cmd.run()
            err = stderr.getvalue().split("\n")[-2]
        self.assertEqual(
            err,
            'ripe-atlas report: error: argument measurement_id: "blaaa" '
            'does not appear to be an existent measurement alias.'
        )

    def test_arg_valid_msm_alias(self):
        """User passed a valid measurement alias."""
        path_aliases = "ripe.atlas.tools.helpers.validators.aliases"
        new_aliases = copy.deepcopy(AliasesDB.DEFAULT)
        new_aliases['measurement']['UNITTEST_ALIAS'] = 1234
        with mock.patch(path_aliases, new_aliases):
            path_get = 'ripe.atlas.tools.commands.report.Command._get_results_from_api'
            with mock.patch(path_get) as mock_get:
                mock_get.side_effect = RipeAtlasToolsException
                with self.assertRaises(RipeAtlasToolsException):
                    self.cmd.init_args(["UNITTEST_ALIAS"])
                    self.cmd.run()
                mock_get.assert_called_once_with(1234)

    def test_measurement_failure(self):
        """Testcase where given measurement id doesn't exist."""
        path = 'ripe.atlas.cousteau.AtlasRequest.get'
        with mock.patch(path) as mock_get:
            mock_get.return_value = False, {}
            with self.assertRaises(RipeAtlasToolsException):
                self.cmd.init_args(["--aggregate-by", "country", "1"])
                self.cmd.run()

    def test_no_results(self):
        """Testcase where given measurement id doesn't have any results."""
        path = 'ripe.atlas.cousteau.AtlasRequest.get'
        with mock.patch(path) as mock_get:
            mock_get.side_effect = [
                (True, {})
            ]
            with self.assertRaises(RipeAtlasToolsException):
                self.cmd.init_args(["--aggregate-by", "country", "1"])
                self.cmd.run()

    def test_no_renderer_found(self):
        """Testcase where renderer canoot be founbd from measurement type."""
        path = 'ripe.atlas.cousteau.AtlasRequest.get'
        with mock.patch(path) as mock_get:
            mock_get.side_effect = [
                (True, {})
            ]
            with self.assertRaises(RipeAtlasToolsException):
                self.cmd.init_args(["--aggregate-by", "country", "1"])
                self.cmd.run()

    def test_valid_case_no_aggr(self):
        """Test case where we have result no aggregation."""
        probes = [
            Probe(id=202, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=677, meta_data={
                "country_code": "DE", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=2225, meta_data={
                "country_code": "DE", "asn_v4": 3332, "asn_v6": "4444"}),
            Probe(id=165, meta_data={
                "country_code": "NL", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=1216, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=270, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=579, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=945, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=879, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
        ]

        with capture_sys_output() as (stdout, stderr):
            path = 'ripe.atlas.cousteau.AtlasRequest.get'
            with mock.patch(path) as mock_get:
                mock_get.side_effect = [
                    (True, self.mocked_results)
                ]
                mpath = 'ripe.atlas.tools.helpers.rendering.Probe.get_many'
                with mock.patch(mpath) as mock_get_many:
                    mock_get_many.return_value = probes
                    self.cmd.init_args(["1"])
                    self.cmd.run()
                    self.assertEquals(
                        stdout.getvalue(), self.expected_output_no_aggr
                    )

    def test_valid_case_with_aggr(self):
        """Test case where we have result with aggregation."""
        expected_output = (
            "RTT_MEDIAN: 40-50\n"
            " 20 bytes from probe #202   178.190.51.206  to hsi.cablecom.ch (62.2.16.24): ttl=52 times:40.024,  40.399,  39.29,  \n"
            " 20 bytes from probe #677   78.128.9.202    to hsi.cablecom.ch (62.2.16.24): ttl=54 times:40.715,  40.259,  40.317, \n"
            "RTT_MEDIAN: 10-20\n"
            " 20 bytes from probe #2225  46.126.90.165   to hsi.cablecom.ch (62.2.16.24): ttl=56 times:10.858,  12.632,  20.53,   32.775,  47.509,  62.745,  78.54,   93.272,  109.738,\n"
            "RTT_MEDIAN: 50-100\n"
            " 20 bytes from probe #165   194.85.27.7     to hsi.cablecom.ch (62.2.16.24): ttl=48 times:87.825,  87.611,  91.0,   \n"
            "RTT_MEDIAN: 20-30\n"
            " 20 bytes from probe #1216  109.190.83.40   to hsi.cablecom.ch (62.2.16.24): ttl=54 times:27.429,  25.672,  25.681, \n"
            " 20 bytes from probe #270   188.192.110.111 to hsi.cablecom.ch (62.2.16.24): ttl=51 times:28.527,  26.586,  26.393, \n"
            " 20 bytes from probe #579   195.88.195.170  to hsi.cablecom.ch (62.2.16.24): ttl=51 times:23.201,  22.981,  22.863, \n"
            " 20 bytes from probe #945   92.111.237.94   to hsi.cablecom.ch (62.2.16.24): ttl=56 times:61.665,  23.833,  23.269, \n"
            "RTT_MEDIAN: 30-40\n"
            " 20 bytes from probe #879   94.254.125.2    to hsi.cablecom.ch (62.2.16.24): ttl=53 times:34.32,   34.446,  34.376, \n"
        )
        probes = [
            Probe(id=202, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=677, meta_data={
                "country_code": "DE", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=2225, meta_data={
                "country_code": "DE", "asn_v4": 3332, "asn_v6": "4444"}),
            Probe(id=165, meta_data={
                "country_code": "NL", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=1216, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=270, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=579, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=945, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=879, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
        ]

        with capture_sys_output() as (stdout, stderr):
            path = 'ripe.atlas.cousteau.AtlasRequest.get'
            with mock.patch(path) as mock_get:
                mock_get.side_effect = [
                    (True, self.mocked_results)
                ]
                mpath = 'ripe.atlas.tools.helpers.rendering.Probe.get_many'
                with mock.patch(mpath) as mock_get_many:
                    mock_get_many.return_value = probes
                    self.cmd.init_args(["--aggregate-by", "rtt-median", "1"])
                    self.cmd.run()
                    expected_set = set(expected_output.split("\n"))
                    returned_set = set(stdout.getvalue().split("\n"))
                    self.assertEquals(returned_set, expected_set)

    def test_asns_filter(self):
        """Test case where user specified probe asns filters.."""
        expected_output = (
            "20 bytes from probe #165   194.85.27.7     to hsi.cablecom.ch (62.2.16.24): ttl=48 times:87.825,  87.611,  91.0,   \n"
            "20 bytes from probe #945   92.111.237.94   to hsi.cablecom.ch (62.2.16.24): ttl=56 times:61.665,  23.833,  23.269, \n"
        )

        probes = [
            Probe(id=202, meta_data={
                "country_code": "GR", "asn_v4": 3337, "asn_v6": "4445"}),
            Probe(id=677, meta_data={
                "country_code": "DE", "asn_v4": 3333, "asn_v6": "4444"}),
            Probe(id=2225, meta_data={
                "country_code": "DE", "asn_v4": 3332, "asn_v6": "4444"}),
            Probe(id=165, meta_data={
                "country_code": "NL", "asn_v4": 3334, "asn_v6": "4444"}),
            Probe(id=1216, meta_data={
                "country_code": "GR", "asn_v4": 3335, "asn_v6": "4444"}),
            Probe(id=270, meta_data={
                "country_code": "GR", "asn_v4": 3340, "asn_v6": "4444"}),
            Probe(id=579, meta_data={
                "country_code": "GR", "asn_v4": 3336, "asn_v6": "4444"}),
            Probe(id=945, meta_data={
                "country_code": "GR", "asn_v4": 3334, "asn_v6": "4444"}),
            Probe(id=879, meta_data={
                "country_code": "GR", "asn_v4": 3333, "asn_v6": "4444"}),
        ]

        with capture_sys_output() as (stdout, stderr):
            path = 'ripe.atlas.cousteau.AtlasRequest.get'
            with mock.patch(path) as mock_get:
                mock_get.side_effect = [
                    (True, self.mocked_results)
                ]
                mpath = 'ripe.atlas.tools.helpers.rendering.Probe.get_many'
                with mock.patch(mpath) as mock_get_many:
                    mock_get_many.return_value = probes
                    self.cmd.init_args(["1", "--probe-asns", "3334"])
                    self.cmd.run()
                    self.assertEquals(stdout.getvalue(), expected_output)

    def test_user_agent(self):
        standard = "RIPE Atlas Tools (Magellan) {}".format(__version__)
        tests = {
            standard: standard,
            "Some custom agent": "Some custom agent",
            "Some custom agent\nwith a second line": "Some custom agent",
            "x" * 3000: "x" * 128,
            "Πράκτορας χρήστη": "Πράκτορας χρήστη",
            "이것은 테스트 요원": "이것은 테스트 요원",
        }
        self.assertEqual(self.cmd.user_agent, standard)
        for in_string, out_string in tests.items():
            path = "ripe.atlas.tools.commands.base.open"
            content = mock.mock_open(read_data=in_string)
            with mock.patch(path, content, create=True):
                self.assertEqual(Command().user_agent, out_string)



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from ripe.atlas.tools.helpers.sanitisers import sanitise


class TestSanitisersHelper(unittest.TestCase):

    def test_sanitise(self):

        self.assertEqual("clean", sanitise("clean"))
        for i in list(range(0, 32)) + [127]:
            self.assertEqual("unclean", sanitise("unclean" + chr(i)))

        self.assertEqual(None, sanitise(None))
        self.assertEqual(7, sanitise(7))

    def test_sanitise_with_newline_exception(self):
        self.assertEqual(
            "unc\nlean", sanitise("unc\nlean", strip_newlines=False))
        for i in set(list(range(0, 32)) + [127]).difference({10}):
            self.assertEqual(
                "unc\nlean",
                sanitise("unc\nlean" + chr(i), strip_newlines=False)
            )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from ripe.atlas.sagan import Result as SaganResult
from ripe.atlas.sagan import ResultParseError

from ..probes import Probe
from ..renderers.base import Result as MagellanResult


class SaganSet(object):
    """
    We need something that doesn't take up a lot of memory while it's being
    constructed, but that will also spread out into a handy string when we need
    it to.
    """

    def __init__(self, iterable=None, probes=()):
        self._probes = probes
        self._iterable = iterable

    def __iter__(self):

        sagans = []

        for line in self._iterable:

            # line may be a dictionary (parsed JSON)
            if hasattr(line, "strip"):
                line = line.strip()

            # Break out when there's nothing left
            if not line:
                break

            try:
                sagan = SaganResult.get(
                    line,
                    on_error=SaganResult.ACTION_IGNORE,
                    on_warning=SaganResult.ACTION_IGNORE
                )
                if not self._probes or sagan.probe_id in self._probes:
                    sagans.append(sagan)
                if len(sagans) > 100:
                    for sagan in self._attach_probes(sagans):
                        yield sagan
                    sagans = []
            except ResultParseError:
                pass  # Probably garbage in the file

        for sagan in self._attach_probes(sagans):
            yield sagan

    def __next__(self):
        return iter(self).next()

    def next(self):
        return self.__next__()

    @staticmethod
    def _attach_probes(sagans):
        probes = dict(
            [(p.id, p) for p in Probe.get_many(s.probe_id for s in sagans)]
        )
        for sagan in sagans:
            sagan.probe = probes[sagan.probe_id]
            yield sagan


class Rendering(object):

    def __init__(self, renderer=None, header="", footer="", payload=()):

        self.renderer = renderer
        self.header = header + "\n" if header else ""
        self.footer = footer + "\n" if footer else ""
        self.payload = payload

    def render(self):

        if self.renderer.SHOW_DEFAULT_HEADER:
            print(self.header, end="")

        print(self.renderer.header(), end="")

        self._smart_render(self.payload)

        print(self.renderer.additional(self.payload), end="")

        print(self.renderer.footer(), end="")

        if self.renderer.SHOW_DEFAULT_FOOTER:
            print(self.footer, end="")

    def _get_rendered_results(self, data):
        for sagan in data:
            yield MagellanResult(self.renderer.on_result(sagan), sagan.probe_id)

    def _smart_render(self, data, indent=""):
        """
        Traverses the aggregation data and prints everything nicely indented.
        """

        if not data:
            return

        if isinstance(data, (list, SaganSet)):

            for line in self._get_rendered_results(data):
                print(indent + line, end="")

        elif isinstance(data, dict):

            for k, v in data.items():
                print("{}{}".format(indent, k))
                self._smart_render(v, indent=indent + " ")



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tzlocal import get_localzone

from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_DNS]
    TIME_FORMAT = "%a %b %d %H:%M:%S %Z %Y"

    def on_result(self, result):

        created = result.created.astimezone(get_localzone())
        probe_id = result.probe_id

        r = "\n\nProbe #{0}\n{1}\n".format(probe_id, "=" * 79)
        if result.responses:
            for response in result.responses:
                r += self.get_formatted_response(probe_id, created, response)
        else:
            r += "\n  {}\n".format(colourise("No response found", "red"))

        return r

    @classmethod
    def get_formatted_response(cls, probe_id, created, response):

        if not response.abuf:
            return "\n- {0} -\n\n  No abuf found.\n".format(
                response.response_id)

        header_flags = []
        for flag in ("aa", "ad", "cd", "qr", "ra", "rd",):
            if getattr(response.abuf.header, flag):
                header_flags.append(flag)

        edns = ""
        if response.abuf.edns0:
            edns = "\n  ;; OPT PSEUDOSECTION:\n  ; EDNS: version: {0}, " \
                   "flags:; udp: {1}\n".format(
                       response.abuf.edns0.version,
                       response.abuf.edns0.udp_size
                   )

        question = ""
        if response.abuf.questions:
            question = response.abuf.questions[0].name

        return cls._colourise_by_response(response, cls.render(

            "reports/dns.txt",

            # Older measurements don't have a response_id
            response_id=response.response_id or 1,

            probe=probe_id,

            question_name=sanitise(question),
            header_opcode=response.abuf.header.opcode,
            header_return_code=response.abuf.header.return_code,
            header_id=response.abuf.header.id,
            header_flags=" ".join(header_flags),
            edns=edns,

            question_count=len(response.abuf.questions),
            answer_count=len(response.abuf.answers),
            authority_count=len(response.abuf.authorities),
            additional_count=len(response.abuf.additionals),

            question=sanitise(cls.get_section(
                "question", response.abuf.questions), strip_newlines=False),
            answers=sanitise(cls.get_section(
                "answer", response.abuf.answers), strip_newlines=False),
            authorities=sanitise(cls.get_section(
                "authority", response.abuf.authorities), strip_newlines=False),
            additionals=sanitise(cls.get_section(
                "additional", response.abuf.additionals), strip_newlines=False),

            response_time=response.response_time,
            response_size=response.response_size,
            created=created.strftime(cls.TIME_FORMAT),
            destination_address=sanitise(response.destination_address),

        ))

    @staticmethod
    def get_section(header, data):

        if not data:
            return ""

        return "\n  ;; {0} SECTION:\n{1}\n".format(
            header.upper(),
            "\n".join(["  {0}".format(_) for _ in data])
        )

    @staticmethod
    def _colourise_by_response(response, output):
        colour = "red" if response.is_error else "green"
        return colourise(output, colour)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .exceptions import RipeAtlasToolsException


class FilterFactory(object):

    @staticmethod
    def create(key, value):
        """Create new filter class based on the key"""
        if key == "asn":
            return ASNFilter(value)
        else:
            return Filter(key, value)


class Filter(object):
    """
    Class that represents filter for results. For now supports only attributes
    of probes property of Result property. It could be extended for any property
    of Result easily.
    """

    def __init__(self, key, value):
        self.key = key
        self.value = value

    def filter(self, result):
        """
        Decide if given result should be filtered (False) or remain on the
        pile of results.
        """
        try:
            attr_value = getattr(result.probe, self.key)
        except AttributeError:
            log = (
                "Cousteau's Probe class does not have an attribute "
                "called: <{}>"
            ).format(self.key)
            raise RipeAtlasToolsException(log)
        if attr_value == self.value:
            return True

        return False


class ASNFilter(Filter):
    """Class thar represents filter by probes that belong to given ASN."""

    def __init__(self, value):
        key = "asn"
        super(ASNFilter, self).__init__(key, value)

    def filter(self, result):
        asn_v4 = getattr(result.probe, "asn_v4")
        asn_v6 = getattr(result.probe, "asn_v6")
        if self.value in (asn_v4, asn_v6):
            return True

        return False


def filter_results(filters, results):
    """docstring for filter"""
    new_results = []
    for result in results:
        for rfilter in filters:
            if rfilter.filter(result):
                new_results.append(result)
                break

    return new_results



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_PING]

    def on_result(self, result):

        packets = result.packets

        if not packets:
            return "No packets found\n"

        # Because the origin value is more reliable as "from" in v4 and as
        # "packet.source_address" in v6.
        origin = result.origin
        if ":" in origin:
            origin = packets[0].source_address

        line = "{} bytes from probe #{:<5} {:15} to {} ({}): ttl={} times:{}\n"
        return line.format(
            result.packet_size,
            result.probe_id,
            origin,
            result.destination_name,
            result.destination_address,
            packets[0].ttl,
            " ".join(["{:8}".format(str(_.rtt) + ",") for _ in packets])
        )
